'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _min = require('lodash/min');

var _min2 = _interopRequireDefault(_min);

var _max = require('lodash/max');

var _max2 = _interopRequireDefault(_max);

var _helpers = require('./helpers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * @param {(Object|Object[])} geojson - https://tools.ietf.org/html/rfc7946
 * @param {('square'|'hexagon')} options.shape - default 'square'
 * @param {number} options.width - in metre, default 1000, min 500, max 500000
 * @param {number} options.tilt - in deg, default 0
 * @param {[number, number]} options.center - [lon, lat] of grid origin
 * @param {Object} options.projection
 * @param {Function} options.projection.forward - map lonlat to grid coordinates
 * @param {Function} options.projection.inverse - map grid coordinates to lonlat
 */
module.exports = function (geojson) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  // normalize geojson input
  var input = [];
  function extractPolygons(node) {
    if ((typeof node === 'undefined' ? 'undefined' : _typeof(node)) !== 'object') return;
    if (node instanceof Array) {
      node.forEach(extractPolygons);
    } else if (node.type === 'Polygon') {
      input.push(node.coordinates);
    } else if (node.type === 'MultiPolygon') {
      var _input;

      (_input = input).push.apply(_input, _toConsumableArray(node.coordinates));
    } else if (node.type === 'Feature') {
      extractPolygons(node.geometry);
    } else if (node.type === 'GeometryCollection') {
      node.geometries.forEach(extractPolygons);
    } else if (node.type === 'FeatureCollection') {
      node.features.forEach(extractPolygons);
    }
  }
  extractPolygons(geojson);

  input = input.map(function (coordinates) {
    return {
      coordinates: coordinates,
      bbox: [(0, _min2.default)(coordinates[0].map(function (point) {
        return point[0];
      })), (0, _min2.default)(coordinates[0].map(function (point) {
        return point[1];
      })), (0, _max2.default)(coordinates[0].map(function (point) {
        return point[0];
      })), (0, _max2.default)(coordinates[0].map(function (point) {
        return point[1];
      }))]
    };
  });

  // global bbox
  var bbox = [(0, _min2.default)(input.map(function (polygon) {
    return polygon.bbox[0];
  })), (0, _min2.default)(input.map(function (polygon) {
    return polygon.bbox[1];
  })), (0, _max2.default)(input.map(function (polygon) {
    return polygon.bbox[2];
  })), (0, _max2.default)(input.map(function (polygon) {
    return polygon.bbox[3];
  }))];

  // set default options
  options.shape = options.shape || 'square';
  options.rotate = options.rotate || 0;
  options.width = options.width || 1000;
  options.width = Math.max(options.width, 500);
  options.width = Math.min(options.width, 500000);
  options.center = options.center || [(bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2];
  options.projection = options.projection || (0, _helpers.equirectangular)(options.center, options.width);

  var forward = options.projection.forward;
  var inverse = options.projection.inverse;
  // different grid spacing for hexagon
  var step = options.shape === 'hexagon' ? Math.sqrt(3) / 4 : 1;
  var grid = {};
  var output = [];

  /*
    given grid orientation (beta) and a set of potential endpoints,
    find min and max grid number (not inclusive of extreme values)
  */
  function dRange(beta, endpoints) {
    var dValues = endpoints.map(function (ep) {
      return (0, _helpers.dotProduct)(beta, ep);
    });
    return {
      min: Math.floor((0, _min2.default)(dValues) / step + 1),
      max: Math.ceil((0, _max2.default)(dValues) / step - 1)
    };
  }

  var corners = [forward([bbox[0], bbox[1]]), forward([bbox[2], bbox[3]]), forward([bbox[0], bbox[3]]), forward([bbox[2], bbox[1]])];

  if (options.shape === 'square') {
    (function () {
      var beta0 = (0, _helpers.polar2cartesian)(options.tilt);
      var beta1 = (0, _helpers.polar2cartesian)(options.tilt + 90);
      var dRange0 = dRange(beta0, corners);
      var dRange1 = dRange(beta1, corners);

      // enumerate through all potential grid cell
      for (var i = dRange0.min - 1; i <= dRange0.max + 1; i++) {
        grid[i] = {};
        for (var j = dRange1.min - 1; j <= dRange1.max + 1; j++) {
          grid[i][j] = {};
        }
      }

      input.forEach(function (polygon) {
        polygon.coordinates.forEach(function (linearRing) {
          linearRing = linearRing.map(forward);
          for (var n = 0; n < linearRing.length - 1; n++) {
            /*
              If a line segment of the input polygon cuts a grid line,
              label the cell directly above and below the point
              where the line segment cuts the grid line 'keep'.
              This traces out the grid cells lying on the edge of the input polygon
            */
            var beta = [linearRing[n + 1][1] - linearRing[n][1], linearRing[n][0] - linearRing[n + 1][0]];
            var d = linearRing[n][0] * linearRing[n + 1][1] - linearRing[n][1] * linearRing[n + 1][0];

            var iRange = dRange(beta0, [linearRing[n], linearRing[n + 1]]);
            var iIntersection = (0, _helpers.linearSolver)(beta0, beta);
            for (var _i2 = iRange.min; _i2 <= iRange.max; _i2++) {
              var intersection = iIntersection(_i2, d);
              var _j2 = Math.floor((0, _helpers.dotProduct)(beta1, intersection));
              grid[_i2][_j2].keep = true;
              grid[_i2 - 1][_j2].keep = true;
            }

            var jRange = dRange(beta1, [linearRing[n], linearRing[n + 1]]);
            var jIntersection = (0, _helpers.linearSolver)(beta1, beta);
            for (var _j3 = jRange.min; _j3 <= jRange.max; _j3++) {
              var _intersection = jIntersection(_j3, d);
              var _i3 = Math.floor((0, _helpers.dotProduct)(beta0, _intersection));
              grid[_i3][_j3].keep = true;
              grid[_i3][_j3 - 1].keep = true;
            }
          }
        });
      });

      // translate grid cells into output polygons
      var getIntersection = (0, _helpers.linearSolver)(beta0, beta1);

      for (var _i in grid) {
        var _i4 = +_i;
        for (var _j in grid[_i4]) {
          var _j4 = +_j;
          output.push({
            id: [_i4, _j4].join('.').replace(/-/g, 'M'),
            type: 'Feature',
            properties: {
              address: [_i4, _j4],
              center: inverse(getIntersection(_i4 + 0.5, _j4 + 0.5))
            },
            geometry: {
              type: 'Polygon',
              coordinates: [[inverse(getIntersection(_i4, _j4)), inverse(getIntersection(_i4 + 1, _j4)), inverse(getIntersection(_i4 + 1, _j4 + 1)), inverse(getIntersection(_i4, _j4 + 1))]]
            },
            keep: grid[_i4][_j4].keep
          });
        }
      }
    })();
  } else if (options.shape === 'hexagon') {
    (function () {
      var beta0 = (0, _helpers.polar2cartesian)(options.tilt);
      var beta1 = (0, _helpers.polar2cartesian)(options.tilt + 60);
      var beta2 = (0, _helpers.polar2cartesian)(options.tilt + 120);
      var dRange0 = dRange(beta0, corners);
      var dRange1 = dRange(beta1, corners);

      for (var i = dRange0.min - 2; i <= dRange0.max + 2; i++) {
        grid[i] = {};
        for (var j = dRange1.min - 2; j <= dRange1.max + 2; j++) {
          var _grid$i$j;

          grid[i][j] = (_grid$i$j = {}, _defineProperty(_grid$i$j, j - i + 1, {}), _defineProperty(_grid$i$j, j - i - 1, {}), _grid$i$j);
        }
      }

      input.forEach(function (polygon) {
        polygon.coordinates.forEach(function (linearRing) {
          linearRing = linearRing.map(forward);
          for (var n = 0; n < linearRing.length - 1; n++) {
            // similar logic as above
            var beta = [linearRing[n + 1][1] - linearRing[n][1], linearRing[n][0] - linearRing[n + 1][0]];
            var d = linearRing[n][0] * linearRing[n + 1][1] - linearRing[n][1] * linearRing[n + 1][0];

            var iRange = dRange(beta0, [linearRing[n], linearRing[n + 1]]);
            var iIntersection = (0, _helpers.linearSolver)(beta0, beta);
            for (var _i5 = iRange.min; _i5 <= iRange.max; _i5++) {
              var intersection = iIntersection(_i5 * step, d);
              var _j5 = Math.floor((0, _helpers.dotProduct)(beta1, intersection) / step);
              var k = Math.floor((0, _helpers.dotProduct)(beta2, intersection) / step);
              if (_i5 - _j5 + k === 1 || _i5 - _j5 + k === -1) {
                grid[_i5][_j5][k].keep = true;
                grid[_i5][_j5 + 1][k + 1].keep = true;
              } else {
                grid[_i5][_j5 + 1][k].keep = true;
                grid[_i5][_j5][k + 1].keep = true;
              }
            }

            var jRange = dRange(beta1, [linearRing[n], linearRing[n + 1]]);
            var jIntersection = (0, _helpers.linearSolver)(beta1, beta);
            for (var _j6 = jRange.min; _j6 <= jRange.max; _j6++) {
              var _intersection2 = jIntersection(_j6 * step, d);
              var _i6 = Math.floor((0, _helpers.dotProduct)(beta0, _intersection2) / step);
              var _k = Math.floor((0, _helpers.dotProduct)(beta2, _intersection2) / step);
              if (_i6 - _j6 + _k === 1 || _i6 - _j6 + _k === -1) {
                grid[_i6][_j6][_k].keep = true;
                grid[_i6 + 1][_j6][_k + 1].keep = true;
              } else {
                grid[_i6 + 1][_j6][_k].keep = true;
                grid[_i6][_j6][_k + 1].keep = true;
              }
            }

            var kRange = dRange(beta2, [linearRing[n], linearRing[n + 1]]);
            var kIntersection = (0, _helpers.linearSolver)(beta2, beta);
            for (var _k2 = kRange.min; _k2 <= kRange.max; _k2++) {
              var _intersection3 = kIntersection(_k2 * step, d);
              var _i7 = Math.floor((0, _helpers.dotProduct)(beta0, _intersection3) / step);
              var _j7 = Math.floor((0, _helpers.dotProduct)(beta1, _intersection3) / step);
              if (_i7 - _j7 + _k2 === 1 || _i7 - _j7 + _k2 === -1) {
                grid[_i7][_j7][_k2].keep = true;
                grid[_i7 + 1][_j7 + 1][_k2].keep = true;
              } else {
                grid[_i7 + 1][_j7][_k2].keep = true;
                grid[_i7][_j7 + 1][_k2].keep = true;
              }
            }
          }
        });
      });

      var getIntersection = (0, _helpers.linearSolver)(beta0, beta1);

      for (var _i in grid) {
        var _i8 = +_i;
        if (!grid[_i8 - 1] || !grid[_i8 + 1]) continue;
        for (var _j in grid[_i8]) {
          var _j8 = +_j;
          if (!grid[_i8][_j8 - 1] || !grid[_i8][_j8 + 1]) continue;
          if ((_i8 + _j8) % 3 === 0) {
            // combines six triangular grid cells into one hexagon grid cell
            var k = _j8 - _i8;
            output.push({
              id: [_i8, _j8].join('.').replace(/-/g, 'M'),
              type: 'Feature',
              properties: {
                address: [_i8, _j8],
                center: inverse(getIntersection(_i8 * step, _j8 * step))
              },
              geometry: {
                type: 'Polygon',
                coordinates: [[inverse(getIntersection(_i8 * step, (_j8 + 1) * step)), inverse(getIntersection((_i8 - 1) * step, _j8 * step)), inverse(getIntersection((_i8 - 1) * step, (_j8 - 1) * step)), inverse(getIntersection(_i8 * step, (_j8 - 1) * step)), inverse(getIntersection((_i8 + 1) * step, _j8 * step)), inverse(getIntersection((_i8 + 1) * step, (_j8 + 1) * step))]]
              },
              keep: [[_i8, _j8, k + 1], [_i8 - 1, _j8, k], [_i8, _j8 - 1, k], [_i8, _j8, k - 1], [_i8 + 1, _j8, k], [_i8, _j8 + 1, k]].some(function (_ref) {
                var _ref2 = _slicedToArray(_ref, 3),
                    i = _ref2[0],
                    j = _ref2[1],
                    k = _ref2[2];

                return grid[i][j][k].keep;
              })
            });
          }
        }
      }
    })();
  }

  output = output.filter(function (feature) {
    if (feature.keep) return true;
    /*
      if center of grid cell lies inside one of the input polygons,
      include that grid cell in final output
    */
    var center = feature.properties.center;
    return input.some(function (polygon) {
      if (center[0] < polygon.bbox[0]) return false;
      if (center[1] < polygon.bbox[1]) return false;
      if (center[0] > polygon.bbox[2]) return false;
      if (center[1] > polygon.bbox[3]) return false;

      var _polygon$coordinates = _toArray(polygon.coordinates),
          outerRing = _polygon$coordinates[0],
          innerRings = _polygon$coordinates.slice(1);

      return (0, _helpers.isInside)(center, outerRing) && innerRings.every(function (innerRing) {
        return !(0, _helpers.isInside)(center, innerRing);
      });
    });
  });

  output.forEach(function (feature) {
    delete feature.keep;
    // copy first point to complete linearRing
    var linearRing = feature.geometry.coordinates[0];
    linearRing.push(linearRing[0]);
  });

  return output;
};